/* Copyright (C) 2004 - 2005  db4objects Inc.  http://www.db4o.com

This file is part of the db4o open source object database.

db4o is free software; you can redistribute it and/or modify it under
the terms of version 2 of the GNU General Public License as published
by the Free Software Foundation and as clarified by db4objects' GPL 
interpretation policy, available at
http://www.db4o.com/about/company/legalpolicies/gplinterpretation/
Alternatively you can write to db4objects, Inc., 1900 S Norfolk Street,
Suite 350, San Mateo, CA 94403, USA.

db4o is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. */
package com.db4o.test;

import com.db4o.*;
import com.db4o.query.*;

public class MassUpdates {
    
    public void storeOne(){
        Test.deleteAllInstances(MassUpdateHelper.class);
        for (int i = 0; i < 5; i++) {
            Test.store(new MassUpdateHelper("muh" + i));
        }
    }
    
    public void testOne(){
        
        Query q = Test.query();
        
        // simple mass update
        q.constrain(MassUpdateHelper.class);
        q.constrain(new Evaluation() {
            public void evaluate(Candidate candidate) {
                ObjectContainer objectContainer = candidate.objectContainer();
                MassUpdateHelper muh = (MassUpdateHelper)candidate.getObject();
                muh.name = "update1";
                objectContainer.set(muh);
                objectContainer.commit();
            }
        });
        q.execute();
        
        check("update1");
        
        
        // triggering mass updates with a singleton
        // complete server-side execution
        q = Test.query();
        q.constrain(this.getClass());
        q.constrain(new Evaluation() {
            public void evaluate(Candidate candidate) {
                ObjectContainer objectContainer = candidate.objectContainer();
                Query q2 = objectContainer.query();
                q2.constrain(MassUpdateHelper.class);
                ObjectSet objectSet = q2.execute();
                while(objectSet.hasNext()){
                    MassUpdateHelper muh = (MassUpdateHelper)objectSet.next();
                    muh.name = "update2";
                    objectContainer.set(muh);
                }
                objectContainer.commit();
            }
        });
        q.execute();
        
        check("update2");
    }
    
    private void check(String name){
        ObjectContainer objectContainer = Test.objectContainer();
        Query q = Test.query();
        q.constrain(MassUpdateHelper.class);
        ObjectSet objectSet = q.execute();
        while(objectSet.hasNext()){
            MassUpdateHelper muh = (MassUpdateHelper)objectSet.next();
            
            // make sure that all objects are reloaded, so we get 
            // no local caching side-effects in C/S mode
            objectContainer.deactivate(muh, Integer.MAX_VALUE);
            objectContainer.activate(muh, Integer.MAX_VALUE);
            Test.ensure(muh.name.equals(name));
        }
    }
    
    
    
    public static class MassUpdateHelper{
        
        public String name;
        
        public MassUpdateHelper(){
        }
        
        public MassUpdateHelper(String name){
            this.name = name;
        }
        
    }
}
